// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "bhv_defender_kick.h"

#include "body_kick_to_corner.h"

#include <rcsc/action/body_advance_ball.h>
#include <rcsc/action/body_dribble.h>
#include <rcsc/action/body_hold_ball.h>
#include <rcsc/action/body_pass.h>
#include <rcsc/action/body_smart_kick.h>
#include <rcsc/action/body_clear_ball.h>
#include <rcsc/action/neck_scan_field.h>
#include <rcsc/action/neck_turn_to_low_conf_teammate.h>

#include <rcsc/player/player_agent.h>
#include <rcsc/player/debug_client.h>

#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>
#include <rcsc/geom/sector_2d.h>
#include "original/body_kubo_dribble.h"
#include "strategy.h"
#include "original/offense_system.h"

/*-------------------------------------------------------------------*/
/*!

*/
bool
Bhv_DefenderKick::execute( rcsc::PlayerAgent * agent )
{
    rcsc::dlog.addText( rcsc::Logger::TEAM,
                        "%s:%d: Bhv_DefenderKick"
                        ,__FILE__, __LINE__ );

    const rcsc::WorldModel & wm = agent->world();

    //pass near front mate
    if( wm.opponentsFromSelf().empty() )return false;
    double nearest_opp_dist = (wm.opponentsFromSelf().front()->pos()
                               - wm.self().pos()).r();
    const rcsc::PlayerObject nearest_mate = *wm.teammatesFromSelf().front();
    double nearest_mate_dist = (nearest_mate.pos() - wm.self().pos()).r();
    
    if( wm.self().pos().x >= -35 && wm.self().pos().x <= 35
        && nearest_mate.pos().x >= wm.self().pos().x
        && nearest_opp_dist >= 3 && nearest_mate_dist <= 3)
    {
//        if(rcsc::Body_KickMultiStep(nearest_mate.pos(),1.5,
//                                    false).execute(agent)){
        if(rcsc::Body_SmartKick(nearest_mate.pos(),
                                1.5,1.2,1).execute(agent)){
            agent->debugClient().addMessage( "Choi" );
            rcsc::dlog.addText( rcsc::Logger::TEAM,
                                "%s:%d: choi pass"
                                ,__FILE__, __LINE__ );
            agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
            return true;
        }
    }
    
    //ʳϳؽư
    if( wm.self().pos().x > -30 ) return false;


    //ߤưʬ
    
    static int S_hold_count = 0;

    //pass(no stamina)
    rcsc::Vector2D target;
    if( wm.self().stamina() < 2500 &&
        rcsc::Body_Pass::get_best_pass(wm,&target,NULL,NULL) ){
        //ŨǤϥѥʤ
        if( Strategy::get_ball_area( wm.ball().pos() ) !=
            Strategy::BA_ShootChance ||
            Strategy::get_ball_area( target ) ==
            Strategy::BA_ShootChance ){
            rcsc::dlog.addText( rcsc::Logger::TEAM,
                                "%s:%d: no stamina pass"
                                ,__FILE__, __LINE__ );
            rcsc::Body_Pass().execute(agent);
            agent->setNeckAction( new rcsc::Neck_ScanField() );
            S_hold_count = 0;
            return true;
        }
    }
    
    //kuboDribble        
    if( S_hold_count >= 3 && Body_KuboDribble(2,3).execute(agent)){
        agent->setNeckAction( new rcsc::Neck_ScanField() );
        S_hold_count = 0;
        return true;
    }

    //kuboDribble    
    if( Body_KuboDribble(3,3).execute(agent) ){
        agent->setNeckAction( new rcsc::Neck_ScanField() );
        S_hold_count = 0;
        return true;
    }

    //pass
    if( rcsc::Body_Pass::get_best_pass(wm,&target,NULL,NULL) ){
        //ŨǤϥѥʤ
        if( Strategy::get_ball_area( wm.ball().pos() ) !=
            Strategy::BA_ShootChance ||
            Strategy::get_ball_area( target ) ==
            Strategy::BA_ShootChance ){
            rcsc::Body_Pass().execute(agent);
            agent->setNeckAction( new rcsc::Neck_ScanField() );
            S_hold_count = 0;
            return true;
        }
    }

    //clear
    if( Strategy::get_ball_area( wm.ball().pos() ) == Strategy::BA_Danger ){
        if( rcsc::Body_ClearBall().execute(agent) ){
            agent->setNeckAction( new rcsc::Neck_ScanField() );
            S_hold_count = 0;
            return true;
        }
    }

    //avoid near opp
    rcsc::Vector2D min_cong_point = OffenseSystem::getMinCongestionPoint(agent,
                                                                         1.5);
    if( min_cong_point.absX() < 47.0 && min_cong_point.absY() < 30.0 ){
        rcsc::Body_Dribble( min_cong_point,
                            0.5,
                            rcsc::ServerParam::i().maxPower(),
                            2 ).execute( agent );
        agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
        ++S_hold_count;
        return true;
    }

    //hold ball
    rcsc::Body_HoldBall().execute( agent );
    agent->setNeckAction( new rcsc::Neck_ScanField() );
    ++S_hold_count;
    
    return true;
}
