// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "bhv_forward_kick.h"

#include "body_kick_to_corner.h"

#include <rcsc/action/body_advance_ball.h>
#include <rcsc/action/body_dribble.h>
#include <rcsc/action/body_hold_ball.h>
#include <rcsc/action/body_clear_ball.h>
#include <rcsc/action/body_pass.h>
#include <rcsc/action/neck_scan_field.h>
#include <rcsc/action/neck_turn_to_low_conf_teammate.h>

#include <rcsc/player/player_agent.h>
#include <rcsc/player/debug_client.h>

#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>
#include <rcsc/geom/sector_2d.h>

#include <rcsc/action/body_smart_kick.h>

#include "body_one_on_one_dribble.h"
#include "body_zulu.h"
#include "body_kubo_dribble.h"

#include <iostream>

#include "original/learn/global_config.h"
#include "original/learn/state.h"
#include "original/offense_system.h"
#include "strategy.h"

/*-------------------------------------------------------------------*/
/*!

*/
bool
Bhv_ForwardKick::execute( rcsc::PlayerAgent * agent )
{
    rcsc::dlog.addText( rcsc::Logger::TEAM,
                        "%s:%d: Bhv_ForwardKick"
                        ,__FILE__, __LINE__ );

    const rcsc::WorldModel & wm = agent->world();
    
    //pass near front mate
    
    if( ! wm.opponentsFromSelf().empty() ){
        double nearest_opp_dist = (wm.opponentsFromSelf().front()->pos()
                                   - wm.self().pos()).r();
        const rcsc::PlayerObject nearest_mate = *wm.teammatesFromSelf().front();
        double nearest_mate_dist = (nearest_mate.pos() - wm.self().pos()).r();
        
        if( wm.self().pos().x >= -35 && wm.self().pos().x <= 35
            && nearest_mate.pos().x >= wm.self().pos().x
            && nearest_opp_dist >= 3 && nearest_mate_dist <= 3)
        {
            if(rcsc::Body_SmartKick(nearest_mate.pos(),1.5,
                                    1.0, 2).execute(agent)){
                agent->debugClient().addMessage( "Choi" );
                rcsc::dlog.addText( rcsc::Logger::TEAM,
                                    "%s:%d: choi pass"
                                    ,__FILE__, __LINE__ );
                agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
                return true;
            }
        }
    }
    
    if( wm.self().pos().x < 20 || wm.self().pos().absY() > 20 ){ //magic number    
        //zulu
        if( Body_Zulu().execute(agent) ){
            agent->setNeckAction( new rcsc::Neck_ScanField() );
            return true;
        }
        
        //oneOnOnedribble
        if( Body_OneOnOneDribble().execute(agent) ){
            agent->setNeckAction( new rcsc::Neck_ScanField() );
            return true;
        }
    } else {
        //oneOnOnedribble
        if( Body_OneOnOneDribble().execute(agent) ){
            agent->setNeckAction( new rcsc::Neck_ScanField() );
            return true;
        }
        
        //zulu
        if( Body_Zulu().execute(agent) ){
            agent->setNeckAction( new rcsc::Neck_ScanField() );
            return true;
        }
    }
    
    //ʳϳؽư
//    if( wm.self().pos().x > -30 ) return false;


    //ߤưʬ
    
    static int S_hold_count = 0;
    rcsc::Vector2D pass_target;
    

    //shoot chanceǤư
    static int S_chance_count = 0;
    if( wm.self().pos().x > 32 && wm.self().pos().absY() < 20 ){
        //Ŭ˥塼
        if( wm.self().pos().x > 43.0 ){
            S_chance_count++;
        } else {
            S_chance_count = 0;
        }
        if( S_chance_count > 4 ) //magic number
        {
            rcsc::Rect2D rect( rcsc::Vector2D(43.0,6.0),
                               rcsc::Vector2D(47.0,-6.0) );
            int opp_count = 0;
            const rcsc::PlayerPtrCont & opps = wm.opponentsFromSelf();
            const rcsc::PlayerPtrCont::const_iterator opps_end = opps.end();
            for ( rcsc::PlayerPtrCont::const_iterator it = opps.begin();
                  it != opps_end;
                  ++it )
            {
                if( rect.contains( (*it)->pos() ) )
                    ++opp_count;
            }
            if( opp_count >= 3 && agent->world().getOpponentGoalie()
                 != static_cast< rcsc::PlayerObject * >( 0 )){
                rcsc::Vector2D target_point(rcsc::ServerParam::i().pitchHalfLength(),
                                            0.0 );
                const rcsc::PlayerObject * goalie
                    = agent->world().getOpponentGoalie();
                
                if( goalie->pos().y > 0 )
                    target_point.y = -1 * rcsc::ServerParam::i().goalWidth() /2*0.8;
                else
                    target_point.y = rcsc::ServerParam::i().goalWidth() / 2 * 0.8;
//                rcsc::Body_KickMultiStep(target_point,
//                                   rcsc::ServerParam::i().ballSpeedMax()
//                    ).execute(agent);
                rcsc::Body_SmartKick(target_point,
                                     rcsc::ServerParam::i().ballSpeedMax(),
                                     2.5,
                                     2).execute(agent);
                agent->debugClient().addMessage( "tekito" );
                rcsc::dlog.addText( rcsc::Logger::TEAM,
                                    "%s:%d: !!tekito shoot!!"
                                    ,__FILE__, __LINE__ );
                agent->setNeckAction( new rcsc::Neck_ScanField() );
                return true;
            }           
        }
        
        //kuboDribble
        if( Body_KuboDribble(2,1).execute(agent)){
            agent->setNeckAction( new rcsc::Neck_ScanField() );
            S_hold_count = 0;
            return true;
        }
        //pass
        if( rcsc::Body_Pass::get_best_pass(wm,&pass_target,NULL,NULL) ){
            if( pass_target.x > 32 && pass_target.absY() < 10.5 ){
                rcsc::Body_Pass().execute(agent);
                agent->setNeckAction( new rcsc::Neck_ScanField() );
                S_hold_count = 0;
                return true;
            }
        }    
        //avoid near opp
        rcsc::Vector2D min_cong_point
            = OffenseSystem::getMinCongestionPoint(agent,1.5);
        if( S_hold_count > 3 ){
            rcsc::Body_Dribble( min_cong_point,
                                0.5,
                                rcsc::ServerParam::i().maxPower(),
                                2 ).execute( agent );
            agent->debugClient().addMessage( "avoid" );
            agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
            ++S_hold_count;
            return true;
        }
    }

    //kuboDribble    
    if( Body_KuboDribble(3,2).execute(agent) ){
        agent->setNeckAction( new rcsc::Neck_ScanField() );
        S_hold_count = 0;
        return true;
    }

    //pass
    if( rcsc::Body_Pass::get_best_pass(wm,&pass_target,NULL,NULL) ){
        //ŨǤϥѥʤ
        if( Strategy::get_ball_area( wm.ball().pos() ) !=
            Strategy::BA_ShootChance ||
            Strategy::get_ball_area( pass_target ) ==
            Strategy::BA_ShootChance ){
            rcsc::Body_Pass().execute(agent);
            agent->setNeckAction( new rcsc::Neck_ScanField() );
            S_hold_count = 0;
            return true;
        }
    }

    //clear
    if( Strategy::get_ball_area( wm.ball().pos() ) == Strategy::BA_Danger ){
        if( rcsc::Body_ClearBall().execute(agent) ){
            agent->setNeckAction( new rcsc::Neck_ScanField() );
            S_hold_count = 0;
            return true;
        }
    }
    
    //avoid near opp
    rcsc::Vector2D min_cong_point = OffenseSystem::getMinCongestionPoint(agent,
                                                                         1.5);
    if( Strategy::get_ball_area( wm.ball().pos() ) != Strategy::BA_ShootChance
        && min_cong_point.absX() < 47.0 && min_cong_point.absY() < 33.0 ){
        rcsc::Body_Dribble( min_cong_point,
                            0.5,
                            rcsc::ServerParam::i().maxPower(),
                            2 ).execute( agent );
        agent->debugClient().addMessage( "avoid" );
        agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
        ++S_hold_count;
        return true;
    }

    //hold ball
    rcsc::Body_HoldBall().execute( agent );
    agent->setNeckAction( new rcsc::Neck_ScanField() );
    ++S_hold_count;
    return true;
    

    return true;
}
