// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "bhv_midfielder_kick.h"

#include "body_kick_to_corner.h"

#include <rcsc/action/body_advance_ball.h>
#include <rcsc/action/body_dribble.h>
#include <rcsc/action/body_hold_ball.h>
#include <rcsc/action/body_pass.h>
#include <rcsc/action/body_smart_kick.h>
#include <rcsc/action/body_clear_ball.h>
#include <rcsc/action/neck_scan_field.h>
#include <rcsc/action/neck_turn_to_low_conf_teammate.h>

#include <rcsc/player/player_agent.h>
#include <rcsc/player/debug_client.h>

#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>
#include <rcsc/geom/sector_2d.h>

#include "original/body_one_on_one_dribble.h"
#include "original/body_zulu.h"
#include "original/body_kubo_dribble.h"
#include "original/offense_system.h"
#include "strategy.h"

/*-------------------------------------------------------------------*/
/*!

*/
bool
Bhv_MidfielderKick::execute( rcsc::PlayerAgent * agent )
{
    rcsc::dlog.addText( rcsc::Logger::TEAM,
                        "%s:%d: Bhv_MidfielderKick"
                        ,__FILE__, __LINE__ );

    const rcsc::WorldModel & wm = agent->world();

    //pass near front mate
    if( !wm.opponentsFromSelf().empty() ){
        double nearest_opp_dist = (wm.opponentsFromSelf().front()->pos()
                                   - wm.self().pos()).r();
        const rcsc::PlayerObject nearest_mate = *wm.teammatesFromSelf().front();
        double nearest_mate_dist = (nearest_mate.pos() - wm.self().pos()).r();
        
        if( wm.self().pos().x >= -35 && wm.self().pos().x <= 35
            && nearest_mate.pos().x >= wm.self().pos().x
            && nearest_opp_dist >= 3 && nearest_mate_dist <= 3)
        {
//            if(rcsc::Body_KickMultiStep(nearest_mate.pos(),1.5,
//                                        false).execute(agent)){
            if(rcsc::Body_SmartKick(nearest_mate.pos(),
                                    1.5,
                                    1.0,
                                    1).execute(agent)){
                agent->debugClient().addMessage( "Choi" );
                rcsc::dlog.addText( rcsc::Logger::TEAM,
                                    "%s:%d: choi pass"
                                    ,__FILE__, __LINE__ );
                agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
                return true;
            }
        }
    }
    
    if( wm.self().pos().x < 20 || wm.self().pos().absY() > 20 ){    
        //zulu
        if( Body_Zulu().execute(agent) ){
            agent->setNeckAction( new rcsc::Neck_ScanField() );
            return true;
        }
        
        //oneOnOnedribble
        if( Body_OneOnOneDribble().execute(agent) ){
            agent->setNeckAction( new rcsc::Neck_ScanField() );
            return true;
        }
    } else {
        //oneOnOnedribble
        if( Body_OneOnOneDribble().execute(agent) ){
            agent->setNeckAction( new rcsc::Neck_ScanField() );
            return true;
        }
        
        //zulu
        if( Body_Zulu().execute(agent) ){
            agent->setNeckAction( new rcsc::Neck_ScanField() );
            return true;
        }
    }
    
    //ʳϳؽư
    if( wm.self().pos().x > -30
        && agent->config().playerNumber() == 6 ) return false;
    

    //ߤưʬ
    
    static int S_hold_count = 0;
    bool danger
        = Strategy::get_ball_area( wm.ball().pos() ) == Strategy::BA_Danger;

    //pass(no stamina)
    rcsc::Vector2D pass_target;
    if( !danger && wm.self().stamina() < 2500 &&
        rcsc::Body_Pass::get_best_pass(wm,&pass_target,NULL,NULL) ){
        //ŨǤϥѥʤ
        if( Strategy::get_ball_area( wm.ball().pos() ) !=
            Strategy::BA_ShootChance ||
            Strategy::get_ball_area( pass_target ) ==
            Strategy::BA_ShootChance ){
            rcsc::Body_Pass().execute(agent);
            agent->setNeckAction( new rcsc::Neck_ScanField() );
            S_hold_count = 0;
            return true;
        }
    }


    //shoot chanceǤư
    if( wm.self().pos().x > 32 && wm.self().pos().absY() < 20 ){
        //kuboDribble
        if( Body_KuboDribble(2,1).execute(agent)){
            agent->setNeckAction( new rcsc::Neck_ScanField() );
            S_hold_count = 0;
            return true;
        }
        //pass
        if( rcsc::Body_Pass::get_best_pass(wm,&pass_target,NULL,NULL) ){
            if( pass_target.x > 32 && pass_target.absY() < 10.5 ){
                rcsc::Body_Pass().execute(agent);
                agent->setNeckAction( new rcsc::Neck_ScanField() );
                S_hold_count = 0;
                return true;
            }
        }    
        //avoid near opp
        rcsc::Vector2D min_cong_point
            = OffenseSystem::getMinCongestionPoint(agent,1.5);
        if( S_hold_count > 3 ){
            rcsc::Body_Dribble( min_cong_point,
                                0.5,
                                rcsc::ServerParam::i().maxPower(),
                                2 ).execute( agent );
            agent->debugClient().addMessage( "avoid" );
            agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
            ++S_hold_count;
            return true;
        }
    }            


    //kuboDribble    
    if( Body_KuboDribble(3,2).execute(agent) ){
        agent->setNeckAction( new rcsc::Neck_ScanField() );
        S_hold_count = 0;
        return true;
    }

    //pass
    if( rcsc::Body_Pass::get_best_pass(wm,&pass_target,NULL,NULL) ){
        //ŨǤϥѥʤ
        if( Strategy::get_ball_area( wm.ball().pos() ) !=
            Strategy::BA_ShootChance ||
            Strategy::get_ball_area( pass_target ) ==
            Strategy::BA_ShootChance ){
            rcsc::Body_Pass().execute(agent);
            agent->setNeckAction( new rcsc::Neck_ScanField() );
            S_hold_count = 0;
            return true;
        }
    }

    //clear
    if( danger ){
        if( rcsc::Body_ClearBall().execute(agent) ){
            agent->setNeckAction( new rcsc::Neck_ScanField() );
            S_hold_count = 0;
            return true;
        }
    }

    //avoid near opp
    rcsc::Vector2D min_cong_point = OffenseSystem::getMinCongestionPoint(agent,
                                                                         1.5);
    if( Strategy::get_ball_area( wm.ball().pos() ) != Strategy::BA_ShootChance
        && min_cong_point.absX() < 47.0 && min_cong_point.absY() < 33.0 ){
        rcsc::Body_Dribble( min_cong_point,
                            0.5,
                            rcsc::ServerParam::i().maxPower(),
                            2 ).execute( agent );
        agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
        ++S_hold_count;
        return true;
    }
    
    //hold ball
    rcsc::Body_HoldBall().execute( agent );
    agent->setNeckAction( new rcsc::Neck_ScanField() );
    ++S_hold_count;
    
    return true;
}
