// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "bhv_pass_k.h"

#include <rcsc/action/neck_turn_to_low_conf_teammate.h>
#include <rcsc/action/neck_scan_field.h>

#include <rcsc/player/player_agent.h>
#include <rcsc/player/debug_client.h>

#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>

#include <rcsc/geom/vector_2d.h>

#include <rcsc/action/obsolete/body_kick_multi_step.h>
#include <rcsc/action/body_hold_ball.h>

#include <rcsc/player/say_message_builder.h>
#include <rcsc/player/free_message.h>

#include <sstream>
#include <iostream>
#include "intention_pass.h"

/*-------------------------------------------------------------------*/
/*!
  execute action
*/
bool
Bhv_PassK::execute( rcsc::PlayerAgent * agent )
{
    const rcsc::WorldModel & wm = agent->world();
    const rcsc::PlayerPtrCont & mates = wm.teammatesFromSelf();

    //matesο+ʬʾM_target¸ߤʤ
    if( mates.size() + 1 < M_target || M_target < 2 ){
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d:!!! invalid M_target !!!"
                            ,__FILE__, __LINE__ );
        std::cerr << "invalid M_target" << std::endl;
        rcsc::Body_HoldBall().execute( agent );
        agent->setNeckAction( new rcsc::Neck_ScanField() );
        return true;
    }

    //źM_target-2
    const rcsc::Vector2D target_pos = mates[M_target-2]->pos();
    const int target_unum = mates[M_target-2]->unum();

    //ѥ᡼
    const double target_dist =  target_pos.dist(wm.self().pos());
    double ball_reach_vel;
    ball_reach_vel = 1.2;
    
    //ܡ뽳
    //kick speedĴ
    double first_speed
        = rcsc::calc_first_term_geom_series_last( ball_reach_vel,
                                                  target_dist,
                                                  rcsc::ServerParam::i().ballDecay() );

    if( first_speed > rcsc::ServerParam::i().ballSpeedMax() )
        first_speed = rcsc::ServerParam::i().ballSpeedMax();
    
    rcsc::dlog.addText( rcsc::Logger::TEAM,
                        "%s:%d:!!! Pass %d !!!"
                        ,__FILE__, __LINE__,
                        M_target);
    agent->debugClient().addMessage( "Bhv_PassK" );
    agent->setIntention
        ( new IntentionPass( target_pos,
                             first_speed,
                             2,true,
                             target_unum,
                             wm.time() ) );
    
    rcsc::Body_KickMultiStep( target_pos,
                              first_speed,
                              false // not enforce
        ).execute( agent );
//    rcsc::Body_HoldBall().execute( agent );

    agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
    
    return true;
}
