#include "body_kubo_shoot.h"

#include <rcsc/player/player_agent.h>
#include <rcsc/player/debug_client.h>
#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>

#include <rcsc/action/body_smart_kick.h>
#include <rcsc/action/body_hold_ball.h>
#include <rcsc/math_util.h>

#include "offense_system.h"
#include <rcsc/player/interception.h>

#include <iostream>

bool
Body_KuboShoot::execute( rcsc::PlayerAgent * agent )
{
    rcsc::dlog.addText( rcsc::Logger::SHOOT,
                        "%s:%d: Bhv_KuboShoot"
                        ,__FILE__, __LINE__ );
    
    const rcsc::WorldModel & wm = agent->world();
    bool shoot_flg = false;

    
    //rcsc::Vector2D target_pos = posGoalPostLeft;
    rcsc::Vector2D target_pos(0.0,0.0);
    target_pos.x = rcsc::ServerParam::i().pitchHalfLength();
    
    double ball_speed = 100*wm.self().kickRate();
    double ball_speed_max =  rcsc::ServerParam::i().ballSpeedMax();

    //ball_speed loop
    for(double i=ball_speed;i<=ball_speed_max;i+=0.4) //magic number?
    {
        
        //target_pos loop
		double ratio = 0.75;
        while(1)
        {
            if( ratio > 0.85 )
            {
                shoot_flg = false;
                break;
            }
            
            target_pos.y = rcsc::ServerParam::i().goalWidth() / 2 * ratio;
            if(checkShoot(agent,target_pos,ball_speed,4))
            {
                shoot_flg = true;
                break;
            }
            
            target_pos.y = -1 * rcsc::ServerParam::i().goalWidth() / 2 * ratio;
            if(checkShoot(agent,target_pos,ball_speed,4))
            {
                shoot_flg = true;
                break;
            }

            ratio += 0.05;

        }
        
        if(shoot_flg)
        {
            if(ball_speed+0.5 >= ball_speed_max)
            {
                ball_speed = ball_speed_max;
            }
            else
            {
                ball_speed = ball_speed + 0.5;
            }
            break;  
        }
    }


    if(shoot_flg)
    {
        rcsc::dlog.addText( rcsc::Logger::SHOOT,
                            "%s:%d: shooted"
                            ,__FILE__, __LINE__ );
        agent->debugClient().addMessage( "kuboShoot" );
        
        return rcsc::Body_SmartKick(target_pos,ball_speed,
                                    0,2).execute(agent);
    }
    else
    {
        rcsc::Body_HoldBall().execute( agent );
        return false;
    }

}


//target_posball_velǥ塼Ȥưå
//count_thrΥȤ׻
//M_cycle_dist;͵
bool
Body_KuboShoot::checkShoot( rcsc::PlayerAgent * agent,
                            const rcsc::Vector2D & target_pos,
                            const double & first_speed,
                            const int & count_thr) const
{
    const rcsc::WorldModel & wm = agent->world();
    const rcsc::PlayerCont & opponents = wm.opponents();
    const rcsc::PlayerType * player_type;
    double kickable_area;
    double fast_cycle = 1000;
    bool flg=false,goal_flg=false;
    int fast_unum = -1;
        
    //opp loop
    for ( rcsc::PlayerCont::const_iterator it = opponents.begin();
          it != opponents.end();
          ++it )
    {
    
        //check pos
        if( (*it).pos().x <= wm.self().pos().x - 10 ){
            continue;
        }
        
        //check posCount
        if ( (*it).posCount() >= count_thr){
            continue;
        }

        player_type = (*it).playerTypePtr();

        //check player_type
        if ( ! player_type ){
            continue;
        }

        //check goalie
        if((*it).goalie()){
            kickable_area = rcsc::ServerParam::i().catchableArea();
        } else {
            kickable_area = player_type->kickableArea();
        }


        //calc intercept cycle
        rcsc::Interception interception(wm.ball().pos(),
                                        (target_pos-wm.ball().pos()).setLengthVector(first_speed));
        double cycle = interception.getReachCycle( (*it).pos(),
                                                   &((*it).vel()),
                                                   ( (*it).bodyCount()==0
                                                     ? &((*it).body())
                                                     : NULL ),
                                                   (*it).posCount(),
                                                   kickable_area,
                                                   player_type->realSpeedMax());

        if(cycle < fast_cycle){
            fast_cycle = cycle;
            fast_unum = (*it).unum();
            flg = true;
        }
        if( (*it).goalie() && (*it).posCount() <= 2 )goal_flg=true;
        
    }
    
    rcsc::dlog.addText( rcsc::Logger::SHOOT,
                        "%s:%d: checkShoot: target=(%f,%f)"
                        ,__FILE__, __LINE__, target_pos.x, target_pos.y);
    rcsc::dlog.addText( rcsc::Logger::SHOOT,
                        "%s:%d: checkShoot: f_cycle=%f, f_unum=%d"
                        ,__FILE__, __LINE__, fast_cycle, fast_unum);
    
    if( flg && goal_flg ){
        double ball_cycle = rcsc::calc_length_geom_series(first_speed,
                                                          (target_pos-wm.ball().pos()).r(),
                                                          rcsc::ServerParam::i().ballDecay());
        rcsc::dlog.addText( rcsc::Logger::SHOOT,
                            "%s:%d: checkShoot: ball_cycle=%f"
                            ,__FILE__, __LINE__, ball_cycle);
        if(ball_cycle == -1){
            ball_cycle = 1000;
        }
	
        if(ball_cycle < fast_cycle - M_cycle_dist){
            return true;
        } else {
            return false;
        }
	
    } else {
        return false;
    }


}
