// -*-c++-*-

/*!
  \file intention_dribble2007.cpp
  \brief queued dribble behavior
*/

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "intention_dribble.h"

#include <rcsc/action/basic_actions.h>
#include <rcsc/action/body_go_to_point.h>
#include <rcsc/action/body_intercept.h>
#include <rcsc/action/neck_scan_field.h>
#include <rcsc/action/neck_turn_to_goalie_or_scan.h>
#include <rcsc/action/view_synch.h>

#include <rcsc/player/audio_sensor.h>
#include <rcsc/player/player_agent.h>
#include <rcsc/player/debug_client.h>
#include <rcsc/player/say_message_builder.h>

#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>
#include <rcsc/geom/line_2d.h>

#define USE_CHANGE_VIEW

/*-------------------------------------------------------------------*/
/*!

*/
bool
IntentionDribble::finished( const rcsc::PlayerAgent* agent )
{
    if ( M_dash_step == 0 )
    {
        rcsc::dlog.addText( rcsc::Logger::DRIBBLE,
                            "%s:%d:!!! finished() check finished. empty queue!!!"
                            ,__FILE__, __LINE__ );
        return true;
    }

    // NOTE: use operator+  operator- has bugs
    if ( M_last_execute_time.cycle() + 1 != agent->world().time().cycle() )
    {
        rcsc::dlog.addText( rcsc::Logger::DRIBBLE,
                      "%s:%d:!!! finished(). last execute time is not match!!!"
                      ,__FILE__, __LINE__ );
        return true;
    }

    if ( agent->world().ball().pos().dist2( M_target_point ) < M_dist_thr * M_dist_thr
         && agent->world().self().pos().dist2( M_target_point ) < M_dist_thr * M_dist_thr )
    {
        rcsc::dlog.addText( rcsc::Logger::DRIBBLE,
                      "%s:%d:!!! finished(). reached target point!!!"
                      ,__FILE__, __LINE__ );
        return true;
    }


    // playmode is checked in rcsc::PlayerAgent::parse()
    // and intention queue is totally managed at there.

    rcsc::dlog.addText( rcsc::Logger::DRIBBLE,
                  "%s:%d: finished(). not finished yet."
                  ,__FILE__, __LINE__ );

    return false;
}

/*-------------------------------------------------------------------*/
/*!

*/
bool
IntentionDribble::execute( rcsc::PlayerAgent * agent )
{
    if ( M_dash_step == 0 )
    {
        rcsc::dlog.addText( rcsc::Logger::DRIBBLE,
                      "%s:%d:!!! execute. empty queue.!!!"
                      ,__FILE__, __LINE__ );
        return false;
    }

    const rcsc::WorldModel & wm = agent->world();

    const rcsc::Vector2D ball_next = wm.ball().pos() + wm.ball().vel();
    if ( ball_next.absX() > rcsc::ServerParam::i().pitchHalfLength() - 0.5
         || ball_next.absY() > rcsc::ServerParam::i().pitchHalfWidth() - 0.5 )
    {
        rcsc::dlog.addText( rcsc::Logger::DRIBBLE,
                      "%s:%d:!!! execute(). ball will be out of pitch. stop intention.!!!"
                      ,__FILE__, __LINE__ );
        return false;
    }

    /*--------------------------------------------------------*/
    // execute action
    if ( M_dash_step > 0 )
    {
        if ( ! doDash( agent ) )
        {
            rcsc::dlog.addText( rcsc::Logger::DRIBBLE,
                          "%s:%d:!!! execute() failed to. dash.  clear intention!!!"
                          ,__FILE__, __LINE__ );
            this->clear();
            return false;
        }
    }
    else
    {
        rcsc::dlog.addText( rcsc::Logger::DRIBBLE,
                      "%s:%d:!!! execute(). No command queue!!!"
                      ,__FILE__, __LINE__ );
        this->clear();
        return false;
    }

#ifdef USE_CHANGE_VIEW
    if ( M_dash_step <= 1 )
    {
        agent->setViewAction( new rcsc::View_Synch() );
    }
    else
    {
        agent->setViewAction( new rcsc::View_Normal() );
    }
#endif

    M_last_execute_time = wm.time();

    rcsc::dlog.addText( rcsc::Logger::DRIBBLE,
                  "%s:%d:!!! execute(). done!!!"
                  ,__FILE__, __LINE__ );
    agent->debugClient().addMessage( "IntDribble" );
    agent->debugClient().setTarget( M_target_point );

//     if ( M_dash_step > 0
//          && agent->config().useCommunication() )
//     {
//         rcsc::dlog.addText( rcsc::Logger::ACTION,
//                       __FILE__":!!!  set dribble communication.!!!" );
//         agent->debugClient().addMessage( "SayD" );
//         agent->addSayMessage( new rcsc::DribbleMessage( M_target_point,
//                                                         M_turn_step + M_dash_step ) );
//     }
    
    return true;
}

bool
IntentionDribble::doDash( rcsc::PlayerAgent * agent  )
{

    rcsc::dlog.addText( rcsc::Logger::TEAM,
                        "%s:%d:!!!intention intercept!!!"
                        ,__FILE__, __LINE__ );
    rcsc::Body_Intercept().execute( agent );
    agent->setNeckAction( new rcsc::Neck_ScanField() );
    
    --M_dash_step;

    return true;
}
