// -*-c++-*-

/*!
  \file player_config.cpp
  \brief player configuration Source File
*/

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "learn_config.h"

#include <rcsc/common/logger.h>
#include <rcsc/param/param_map.h>

#include <cassert>
#include <string>

using std::string;

//ִ򤹤ؿ
string replace_all( const string & source,
                    const string & pattern,
                    const string & placement )
{
    string result;
    string::size_type pos_before = 0;
    string::size_type pos = 0;
    string::size_type len = pattern.size();
    while ((pos = source.find(pattern, pos)) != string::npos) {
        result.append(source, pos_before, pos - pos_before);
        result.append(placement);
        pos += len;
        pos_before = pos;
    }
    result.append(source, pos_before, source.size() - pos_before);
    return result;
}


/*-------------------------------------------------------------------*/
/*!

*/
LearnConfig::LearnConfig()
{
    setDefaultParam();
}

/*-------------------------------------------------------------------*/
/*!

*/
LearnConfig::~LearnConfig()
{

}

/*-------------------------------------------------------------------*/
/*!

*/
void
LearnConfig::setDefaultParam()
{

    M_epsilon = 0.01;
    M_alpha = 0.125;
    M_value_data_name = "value-%u";
    M_enable_update = true;

    M_goal_reward = 1.0;
    M_offense_reward = 0.0;
    M_catch_reward = -0.2;
    M_out_reward = -0.2;
    M_defense_reward = -0.2;
    
}

/*-------------------------------------------------------------------*/
/*!

*/
void
LearnConfig::createParamMap( rcsc::ParamMap & param_map )
{
    param_map.add()
        ( "epsilon", "", &M_epsilon, "epsilon" )
        ( "alpha", "", &M_alpha, "alpha" )
        ( "value_data_name", "", &M_value_data_name, "tile weight file name" )
        ( "enable_update", "", &M_enable_update, "enable update value" )
        ( "goal_reward", "", &M_goal_reward, "goal reward" )
        ( "offense_reward", "", &M_offense_reward, "offense reward" )
        ( "catch_reward", "", &M_catch_reward, "catch reward" )
        ( "out_reward", "", &M_out_reward, "out reward" )
        ( "defense_reward", "", &M_defense_reward, "defense reward" )
        ;
}

