// -*-c++-*-

/*!
  \file player_config.h
  \brief player configuration Header File
*/

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifndef LEARN_CONFIG_H
#define LEARN_CONFIG_H

#include <string>
#include <rcsc/param/param_map.h>
#include <sstream>

class rcsc::ParamMap;

std::string replace_all( const std::string & source,
                         const std::string & pattern,
                         const std::string & placement );
    
class LearnConfig {
private:

    double M_epsilon;
    double M_alpha;
    std::string M_value_data_name;
    bool M_enable_update;
    
    double M_goal_reward;
    double M_offense_reward;
    double M_catch_reward;
    double M_out_reward;
    double M_defense_reward;


public:

    LearnConfig();
    
    ~LearnConfig();

    void createParamMap( rcsc::ParamMap & param_map );


protected:

    void setDefaultParam();

public:

    double epsilon() const
    {
        return M_epsilon;
    }

    double alpha() const
    {
        return M_alpha;
    }

    const std::string & valueDataName() const
    {
        return M_value_data_name;
    }

    std::string valueDataName( const int & unum ) const
    {
        std::ostringstream oss;
        oss << unum;
        return replace_all( M_value_data_name, "%u", oss.str() );
    }

    bool enableUpdate() const
    {
        return M_enable_update;
    }

    double goalReward() const
    {
        return M_goal_reward;
    }

    double offenseReward() const
    {
        return M_offense_reward;
    }

    double catchReward() const
    {
        return M_catch_reward;
    }

    double outReward() const
    {
        return M_out_reward;
    }

    double defenseReward() const
    {
        return M_defense_reward;
    }

    //黻ҤΥС
    bool operator==(const LearnConfig & obj)
    {
        if( this->M_epsilon == obj.epsilon() &&
            this->M_alpha == obj.alpha()
            )
            return true;

        return false;
    }
};

#endif
