#include "state.h"
#include <iostream>

#include <rcsc/player/player_agent.h>
#include <rcsc/player/debug_client.h>

#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>

#include <cmath>
#include <vector>

#include <rcsc/geom/sector_2d.h>


inline
double
State::operator[](const int & idx) {
    if( idx < 0 || idx >= M_state_dim ){
        std::cerr << "Error: invalid access idx(state.h)" << std::endl;
        return 0.0;
    }
    return M_array[idx];
}

bool
State::setState( const rcsc::WorldModel & wm )
{

    //ʬޤ᤿̣Υܡᤤvector
    std::vector< const rcsc::AbstractPlayerObject* > offense_players;
    
    const rcsc::PlayerPtrCont & opps = wm.opponentsFromBall();
    const rcsc::PlayerPtrCont::const_iterator opps_end = opps.end();
    const rcsc::PlayerPtrCont & mates = wm.teammatesFromBall();

    //ǧǤƤͿ­ʤϽʤ
    if( opps.size() < 1 || mates.size() < CONSIDERABLE_NUM ){
        std::cerr << wm.time() << " unum"
                  << wm.self().unum() << " "
                  << " **WARNING** opps, mates size not correct "
                  << "(state.cpp)" << std::endl;
        return false;
    }

    bool self_push_flg = false;
    const rcsc::PlayerPtrCont::const_iterator mates_end = mates.end();
    for ( rcsc::PlayerPtrCont::const_iterator it = mates.begin();
          it != mates_end;
          ++it )
    {
        if( !self_push_flg &&
            wm.ball().pos().dist( wm.self().pos() )
            < wm.ball().pos().dist( (*it)->pos() ) )
        {
            offense_players.push_back( &wm.self() );
            self_push_flg = true;
            if( offense_players.size() == CONSIDERABLE_NUM ) break;
        }
        
        
        offense_players.push_back( (*it) );
        if( offense_players.size() == CONSIDERABLE_NUM ) break;
    }
    
    if( offense_players.size() != CONSIDERABLE_NUM ){
        std::cerr << wm.time() << " unum"
                  << wm.self().unum() << " "
                  << " **WARNING** offense_players size not correct "
                  << "(state.cpp)" << std::endl;
        return false;
    }

    //O1Ϥ褯ȤäƤ
    const rcsc::AbstractPlayerObject* offense1 = offense_players.front();
    
    static const
        rcsc::Vector2D right_post(rcsc::ServerParam::i().pitchHalfLength(),
                                  -rcsc::ServerParam::i().goalHalfWidth());
    static const
        rcsc::Vector2D left_post(rcsc::ServerParam::i().pitchHalfLength(),
                                 rcsc::ServerParam::i().goalHalfWidth());
    
    //ե󥹥ץ쥤䤴Ȥˤѿη׻
    
    //ե󥹥ץ쥤Υܡ˶ᤤֹ(ֶᤤΤ1)
    int offense_player_number = 1;
     
    const std::vector< const rcsc::AbstractPlayerObject* >::const_iterator
        op_end = offense_players.end();
    for ( std::vector< const rcsc::AbstractPlayerObject* >::const_iterator
              it = offense_players.begin();
          it != op_end;
          ++it )
    {
        
        //calc min_dist_o_d, min_ang_o_o1_d
        double min_dist=100.0, min_ang=180.0;
        //ǥե󥹥ץ쥤롼
        for ( rcsc::PlayerPtrCont::const_iterator it2 = opps.begin();
              it2 != opps_end;
              ++it2 )
        {
            //goalieФ
            if( (*it2)->goalie() ) continue;

            //min_dist
            if( min_dist > (*it)->pos().dist( (*it2)->pos() ) )
                min_dist = (*it)->pos().dist( (*it2)->pos() );

            if( offense_player_number != 1 ){
                //min_angle
                rcsc::Vector2D o1_to_offense = (*it)->pos() - offense1->pos();
                rcsc::Vector2D o1_to_defense = (*it2)->pos() - offense1->pos();
                double ang
                    = std::acos(o1_to_offense.innerProduct( o1_to_defense )
                                / ( o1_to_offense.r() * o1_to_defense.r() ));
                ang = rcsc::AngleDeg::rad2deg( ang );
                if( min_ang > ang ){
                    min_ang = ang;
                }
            }
            
        }
        
        //calc dist_o_gl
        if( (*it)->pos().y < -rcsc::ServerParam::i().goalHalfWidth() )
            M_dist_o_gl[offense_player_number-1]
                = right_post.dist( (*it)->pos() );
        else if( (*it)->pos().y > rcsc::ServerParam::i().goalHalfWidth() )
            M_dist_o_gl[offense_player_number-1]
                = left_post.dist( (*it)->pos() );
        else
            M_dist_o_gl[offense_player_number-1]
                = std::abs( rcsc::ServerParam::i().pitchHalfLength()
                            - (*it)->pos().x);

        //min_dist_o_d
        M_min_dist_o_d[offense_player_number-1] = min_dist;
        
        if( offense_player_number != 1 ){
            //calc dist_o1_o
            M_dist_o1_o[offense_player_number-2]
                = offense1->pos().dist( (*it)->pos() );
            //min_ang_o_o1_d
            M_min_ang_o_o1_d[offense_player_number-2] = min_ang;
        }

        ++offense_player_number;
    }
     
    //O1ͭطѿη׻
     
    //calc dist_o1_dg
   const rcsc::PlayerObject * goalie = wm.getOpponentGoalie();
   if( goalie ){
       M_dist_o1_dg =  goalie->pos().dist( offense1->pos() );
   }
    
    //calc min_dist_o1_d
    double o1_min_dist = 100.0;
    static const double cone_half_angle = 60.0;
    static const rcsc::Vector2D
        goal_center(rcsc::ServerParam::i().pitchHalfLength(),0);

    //cone_half_angleγ٤Υ˸̤
    const rcsc::Vector2D o1_to_goal
        = goal_center - offense1->pos();
    rcsc::Sector2D dribble_corn( offense1->pos(),
                                 0,
                                 50,
                                 o1_to_goal.th() - cone_half_angle,
                                 o1_to_goal.th() + cone_half_angle );
    //θ̤˴ޤޤŨκǾΥõ
    for ( rcsc::PlayerPtrCont::const_iterator it = opps.begin();
          it != opps_end;
          ++it )
    {
        if ( dribble_corn.contains( (*it)->pos() )
             && o1_min_dist > offense1->pos().dist( (*it)->pos() ) )
        {
            o1_min_dist = offense1->pos().dist( (*it)->pos() );
        }
    }
    M_min_dist_o1_dcone = o1_min_dist;


    
    //calc max_goal_ang_o1
    //ʬȥݥȤֻѷ
    rcsc::Triangle2D goal_corn( offense1->pos(),
                                right_post,
                                left_post );

    //ѷ˴ޤޤŨγ٤
    std::vector< double > opp_angle;

    for ( rcsc::PlayerPtrCont::const_iterator it = opps.begin();
          it != opps_end;
          ++it )
    {
        if ( goal_corn.contains( (*it)->pos() ) )
        {
            opp_angle.push_back( ((*it)->pos()
                                  - offense1->pos()).th().degree() );
        }
    }

    //ݥȤγ٤ɲ
    opp_angle.push_back( (right_post - offense1->pos()).th().degree() );
    opp_angle.push_back( (left_post - offense1->pos()).th().degree() );

    //ǥ
    std::sort(opp_angle.begin(), opp_angle.end());

    //ٺΰ礭Τõ
    double max_diff = 0;
    int max_idx = 0;
    double tmp_diff;
    for(unsigned int i=0; i<opp_angle.size()-1; ++i){
        tmp_diff = opp_angle[i+1] - opp_angle[i];
        if( max_diff < tmp_diff ){
            max_diff = tmp_diff;
            max_idx = i;
        }
    }
    rcsc::AngleDeg max_angle_diff
        = (opp_angle[max_idx+1] - opp_angle[max_idx]);
    M_max_goal_ang_o1 = max_angle_diff.degree();
    

    //֤¸
    M_set_cycle = wm.time().cycle();

    //¸
    std::vector<double> v;
    for(unsigned int i=0; i<CONSIDERABLE_NUM-1; ++i)
        v.push_back( M_dist_o1_o[i] );
    v.push_back( M_dist_o1_dg );
    for(unsigned int i=0; i<CONSIDERABLE_NUM; ++i)
        v.push_back( M_dist_o_gl[i] );
    for(unsigned int i=0; i<CONSIDERABLE_NUM; ++i)
        v.push_back( M_min_dist_o_d[i] );
    for(unsigned int i=0; i<CONSIDERABLE_NUM-1; ++i)
        v.push_back( M_min_ang_o_o1_d[i] );
    v.push_back( M_min_dist_o1_dcone );
    v.push_back( M_max_goal_ang_o1 );

    M_array = v;

    return true;
}

    

void
State::test() const
{
    using namespace std;
    cout << "-----------" << endl;
    cout << "M_dist_o1_o "
         << M_dist_o1_o[0] << " "
         << M_dist_o1_o[1] << " "
         << M_dist_o1_o[2] << " "
         << M_dist_o1_o[3] << " "
         << M_dist_o1_o[4] << " "
         << endl;

    cout << "M_dist_o_gl "
         << M_dist_o_gl[0] << " "
         << M_dist_o_gl[1] << " "
         << M_dist_o_gl[2] << " "
         << M_dist_o_gl[3] << " "
         << M_dist_o_gl[4] << " "
         << M_dist_o_gl[5] << " "
         << endl;
    
    cout << "M_min_dist_o_d "
         << M_min_dist_o_d[0] << " "
         << M_min_dist_o_d[1] << " "
         << M_min_dist_o_d[2] << " "
         << M_min_dist_o_d[3] << " "
         << M_min_dist_o_d[4] << " "
         << M_min_dist_o_d[5] << " "
         << endl;

    cout << "M_min_ang_o_o1_d "
         << M_min_ang_o_o1_d[0] << " "
         << M_min_ang_o_o1_d[1] << " "
         << M_min_ang_o_o1_d[2] << " "
         << M_min_ang_o_o1_d[3] << " "
         << M_min_ang_o_o1_d[4] << " "
         << endl;

    cout << "M_min_dist_o1_d " << M_min_dist_o1_dcone << endl;

    cout << "-----------" << endl;

//     cout << "distO1O "
//          << distO1O(0) << " "
//          << distO1O(1) << " "
//          << distO1O(2) << " "
//          << endl;

//     cout << "distO1Goalie "
//          << distO1Goalie() << endl;
    
//     cout << "distOGoalLine "
//          << distOGoalLine(0) << " "
//          << distOGoalLine(1) << " "
//          << distOGoalLine(2) << " "
//          << distOGoalLine(3) << " "
//          << endl;
    
//     cout << "minDistOD "
//          << minDistOD(0) << " "
//          << minDistOD(1) << " "
//          << minDistOD(2) << " "
//          << minDistOD(3) << " "
//          << endl;

//     cout << "minAngOO1D "
//          << minAngOO1D(0) << " "
//          << minAngOO1D(1) << " "
//          << minAngOO1D(2) << " "
//          << endl;

//     cout << "minDistO1Dcone " << minDistO1Dcone() << endl;

//     cout << "maxGoalAngO1 " << maxGoalAngO1() << endl;



     cout << "-----------" << endl;
//    std::vector<double> vec = getStateList(0,true);
//    for(unsigned int i=0; i<vec.size(); ++i)
//        cout << vec[i] << " ";
//    cout << endl;
    
}
