#ifndef TILE_CODING_H
#define TILE_CODING_H

#include <iostream>
#include <vector>
#include <map>

class TileCoding
{ 
private:

    void print( const std::vector<int> & v ) const
    {
        std::cout << "(";
        for(unsigned int i=0; i<v.size(); ++i){
            if( i )
                std::cout << " " << v[i];
            else
                std::cout << v[i];
        }
        std::cout << ")";
    }

    std::vector< std::vector<int> >
    getTiles( const std::vector<double> & input ) const;
    
    //---ŤߴϢ---
    double getWeight( const std::vector<int> & key ) const
    {
        std::map< std::vector<int>, double>::const_iterator it;
        it = M_weight.find(key);
        
        if( it != M_weight.end() ){
            return it->second;
        } else {
            //Ĥʤ0֤
            return 0.0;
        }
    }
    
    //keyǤdiffѲ
    bool updateWeight( const std::vector<int> & key, const double & diff )
    {
        std::map< std::vector<int>, double>::iterator it;
        it = M_weight.find(key);
        
        if( it != M_weight.end() ){
            it->second += diff;
        } else {
            M_weight[key] = diff;
        }
        return true;
    }
    
private:
    unsigned int M_input_dim;
    int M_tiling_num;
    std::vector<double> M_tile_widths;

    std::map< std::vector<int>, double > M_weight;

public:
    TileCoding(const int & input_dim,
               const int & tiling_num,
               const std::vector<double> & tile_widths);

    double getValue( const std::vector<double> & input ) const;
    bool update( const std::vector<double> & input,
                 const double & diff );

    bool write( const std::string & file_name ) const;
    bool read( const std::string & file_name );

    void test()
    {
        std::map<std::vector<int>, double>::iterator it;
        for(it=M_weight.begin(); it!=M_weight.end(); ++it)
        {
            print(it->first);
            std::cout << " " << it->second << std::endl;
        }
        std::cout << "weight.size() ==  " << M_weight.size() << std::endl;
    }
};


#endif
