#include "offense_system.h"

#include <cmath>

#include <rcsc/action/body_kick_one_step.h>
#include <rcsc/action/body_smart_kick.h>

#include <rcsc/common/server_param.h>
#include <rcsc/common/logger.h>

#include <rcsc/player/interception.h>

int
OffenseSystem::getPointReachCycle(rcsc::PlayerAgent * agent,
                             const rcsc::Vector2D & target_point,
                             const rcsc::AbstractPlayerObject & player)
{
    rcsc::Interception interception( target_point, 0, 0 );

    return static_cast<int>
        ( std::ceil( interception.getReachCycle( player.pos(),
                                                 &(player.vel()),
                                                 ( player.bodyCount()==0
                                                   ? &(player.body())
                                                   : NULL ),
                                                 player.posCount(),
                                                 0.9,
                                                 1.0 ) ) );
}



// bool
// OffenseSystem::simulateKickMultiStep(rcsc::PlayerAgent * agent,
//                                      const rcsc::Vector2D & target_point,
//                                      const double & first_speed,
//                                      rcsc::Vector2D & ball_result_pos,
//                                      rcsc::Vector2D & ball_result_vel,
//                                      int & kick_step)
// {
//     rcsc::Vector2D target_rpos = target_point - agent->world().self().pos();
//     rcsc::Vector2D achieved_vel;

//     if ( rcsc::Body_KickTwoStep::simulate_one_kick( &achieved_vel,
//                                                     NULL,
//                                                     NULL,
//                                                     target_rpos,
//                                                     std::min(first_speed,
//                                                              rcsc::ServerParam::i().ballSpeedMax()),
//                                                     rcsc::Vector2D( 0.0, 0.0 ), // my current pos
//                                                     agent->world().self().vel(),
//                                                     agent->world().self().body(),
//                                                     agent->world().ball().rpos(),
//                                                     agent->world().ball().vel(),
//                                                     agent,
//                                                     false ) ) // not enforce
//     {
//         ball_result_pos = agent->world().ball().pos() + achieved_vel;
//         ball_result_vel = achieved_vel * rcsc::ServerParam::i().ballDecay();;
//         kick_step = 1;

//         return true;
//     }

//     rcsc::Vector2D next_vel;

//     if ( rcsc::Body_KickTwoStep::simulate_two_kick( &achieved_vel,
//                                                     &next_vel,
//                                                     target_rpos,
//                                                     first_speed,
//                                                     rcsc::Vector2D( 0.0, 0.0 ), // my current pos
//                                                     agent->world().self().vel(),
//                                                     agent->world().self().body(),
//                                                     agent->world().ball().rpos(),
//                                                     agent->world().ball().vel(),
//                                                     agent,
//                                                     false ) ) // not enforced
//     {
//         ball_result_pos = agent->world().ball().pos() + next_vel;
//         ball_result_vel = next_vel * rcsc::ServerParam::i().ballDecay();;
//         kick_step = 2;

//         return true;
//     }


//     if ( rcsc::Body_KickMultiStep::simulate_three_kick( &achieved_vel,
//                                                         &next_vel,
//                                                         target_rpos,
//                                                         first_speed,
//                                                         rcsc::Vector2D( 0.0, 0.0 ), // my current pos
//                                                         agent->world().self().vel(),
//                                                         agent->world().self().body(),
//                                                         agent->world().ball().rpos(),
//                                                         agent->world().ball().vel(),
//                                                         agent,
//                                                         false ) )
//     {
//         ball_result_pos = agent->world().ball().pos() + next_vel;
//         ball_result_vel = next_vel * rcsc::ServerParam::i().ballDecay();;
//         kick_step = 3;

//         return true;
//     }


//     return false;



// }



bool
OffenseSystem::pointPredict(rcsc::PlayerAgent * agent,
                            const rcsc::Vector2D & target_point,
                            const int & count_thr,
                            const int & mode,
                            int * fastest_cycle,
                            int * fastest_unum,
                            bool skip_goalie)
{
    const rcsc::WorldModel & wm = agent->world();
    rcsc::PlayerCont mates;
    switch(mode){
    case TEAM:
        mates = wm.teammates();
        break;
    case OPP:
        mates = wm.opponents();
        break;
    default:
        break;
    }

    int cycle,f_cycle=1000,f_unum=-10;
    for ( rcsc::PlayerCont::const_iterator it = mates.begin();
          it != mates.end();
          ++it )
    {
        //skip goalie
        if( skip_goalie && (*it).goalie() ) continue;
        
        //check tackling
        if( (*it).isTackling() ) continue;
        
        //check unum
        if( (*it).unum() == -1 ) continue;
        
        //check count
        if( (*it).posCount() > count_thr ) continue;

        cycle = getPointReachCycle(agent,target_point,(*it));
        if(cycle == 1000) continue;

        if(cycle < f_cycle){
            f_cycle = cycle;
            f_unum = (*it).unum();
        }
    }

    if(f_unum == -10){
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: pointPredict failed."
                            ,__FILE__, __LINE__ );
        return false;
    }
    
    if(fastest_cycle) *fastest_cycle = f_cycle;
    if(fastest_unum) *fastest_unum = f_unum;
    
    return true;
}


double
OffenseSystem::calcCongestion(rcsc::PlayerAgent * agent,
                              const rcsc::Vector2D & target_point)
{
    const rcsc::WorldModel & wm = agent->world();
    double congestion=0;
    
    const rcsc::PlayerPtrCont::const_iterator
        end = wm.opponentsFromSelf().end();
    for( rcsc::PlayerPtrCont::const_iterator
             it = wm.opponentsFromSelf().begin();
         it != end; ++it)
    {
        congestion += 1.0/(*it)->pos().dist2(target_point); 
    }

    return congestion;
}

    
rcsc::Vector2D
OffenseSystem::getMinCongestionPoint(rcsc::PlayerAgent * agent,
                                     const double & r)
{
    const rcsc::WorldModel & wm = agent->world();
    rcsc::Vector2D base_point = rcsc::Vector2D::polar2vector(r,
                                                             wm.self().body());

    //calc congestion
    double congestion[12];
    for(int i=0;i<12;i++)
    {
        rcsc::Vector2D tmp_point
            = wm.self().pos() + base_point.rotate(i * 30.0);
        congestion[i] = calcCongestion(agent,tmp_point);
    }

    //find min
    int min_idx = 0;
    double min_congestion = 100;
    for(int i=0;i<12;i++)
    {
        if( congestion[i] < min_congestion ){
            min_congestion = congestion[i];
            min_idx = i;
        }
    }
    
    rcsc::Vector2D min_point =  wm.self().pos()
        + base_point.rotate(min_idx * 30.0);
    
    rcsc::dlog.addText( rcsc::Logger::TEAM,
                        "%s:%d: min_congestion(%f,%f)=%f"
                        ,__FILE__, __LINE__, min_point.x,
                        min_point.y,congestion[min_idx] );
    
    return wm.self().pos() + base_point.rotate(min_idx * 30.0);
}


double
OffenseSystem::getLineCongestion(rcsc::PlayerAgent * agent,
                                 const double & lineX,
                                 const rcsc::Vector2D & target_point)
{    
    std::vector< rcsc::Vector2D > points;
    std::vector< double > congestions;

    int idx = 0;
    for(int i=-30;i<=30;i+=3){
        points.push_back( rcsc::Vector2D(lineX,(double)i) );
        congestions.push_back( calcCongestion(agent,points[idx]) );
        ++idx;
    }

    int min_idx = 0;
    double min_dist = 100;
    double diff_left,diff_right;
    for(int i=0;i<idx;++i){

        int tmp_idx = i;
        int count = 0;
        while(1){
            //stop inf loop
            if(count++ >= idx) break;
            
            if(tmp_idx != idx-1){
                diff_left = congestions[tmp_idx+1] - congestions[tmp_idx];
            } else {
                diff_left = 0;
            }

            if(tmp_idx != 0){
                diff_right = congestions[tmp_idx-1] - congestions[tmp_idx];
            } else {
                diff_right = 0;
            }
            
            if(diff_left >= 0 && diff_right >= 0) break;

            if(diff_left >= diff_right){
                --tmp_idx;
            } else {
                ++tmp_idx;
            }
        }
        agent->debugClient().addCircle( points[tmp_idx], 1 );
        if( min_dist > target_point.dist(points[tmp_idx]) ){
            min_dist = target_point.dist(points[tmp_idx]);
            min_idx = tmp_idx;
        }
                                         
    }
    
    agent->debugClient().addCircle( points[min_idx], 1.5 );


    return points[min_idx].y;
}
