// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "sample_trainer.h"

#include <rcsc/trainer/trainer_command.h>
#include <rcsc/trainer/trainer_config.h>
#include <rcsc/coach/global_world_model.h>
#include <rcsc/common/basic_client.h>
#include <rcsc/common/player_param.h>
#include <rcsc/common/player_type.h>
#include <rcsc/common/server_param.h>
#include <rcsc/param/param_map.h>
#include <rcsc/param/cmd_line_parser.h>
#include <rcsc/random.h>

#include <iostream>
#include <sstream>
#include <algorithm>
#include <fstream>

#include "original/learn/global_config.h"
#include "original/learn/global.h"

static const int log_freq = 1;

class RandomGenerator
{
private:
public:
    std::ptrdiff_t operator()( std::ptrdiff_t arg ) const
    {
        rcsc::UniformInt uni(0,static_cast<int>(arg)-1);
        // 0 ʾ arg ̤֤
        return static_cast< std::ptrdiff_t >( uni() );
    }
};

/*-------------------------------------------------------------------*/
/*!

 */
SampleTrainer::SampleTrainer()
    : TrainerAgent()
    ,M_init_flg(false)
    ,M_last_decision_time(0,0)
    ,M_episode_count(0)
    ,M_episode_sum(0)
    ,M_goal_count(0)
{

}

/*-------------------------------------------------------------------*/
/*!

*/
SampleTrainer::~SampleTrainer()
{

}

bool
SampleTrainer::outputGoalLog()
{
    //ե
    std::ostringstream goal_log_path;

    goal_log_path << gconfig.exp().logDir() << "/"
                  << gconfig.exp().goalLogName()
                  << ".log";

    std::ofstream ofs( goal_log_path.str().c_str(),
                       std::ios::out | std::ios::app );

    if( ofs ){
        ofs << M_episode_count << " " << M_goal_count << std::endl;
        return true;
    } else {
        return false;
    }
}

//λ
void
SampleTrainer::finalize()
{
//    outputGoalLog();
    rcsc::TrainerAgent::finalize();
}

void
SampleTrainer::handleExit()
{
    finalize();
}


/*-------------------------------------------------------------------*/
/*!

 */
bool
SampleTrainer::initImpl( rcsc::CmdLineParser & cmd_parser )
{
    bool result = rcsc::TrainerAgent::initImpl( cmd_parser );

    rcsc::ParamMap my_params;
#if 0
    std::string formation_conf;
    my_map.add()
        ( &conf_path, "fconf" )
        ;
#endif
    cmd_parser.parse( my_params );

    if ( cmd_parser.count( "help" ) )
    {
        my_params.printHelp( std::cout );
        return false;
    }

    if ( cmd_parser.failed() )
    {
        std::cerr << "trainer: ***WARNING*** detected unsupported options: ";
        cmd_parser.print( std::cerr );
        std::cerr << std::endl;
    }

    if ( ! result )
    {
        return false;
    }

    //////////////////////////////////////////////////////////////////
    // Add your code here.
    //////////////////////////////////////////////////////////////////

    //ѥǥ쥯ȥ
    std::ostringstream oss;
    oss << "if [ ! -d " << gconfig.exp().logDir() << " ]; then mkdir "
        << gconfig.exp().logDir() << "; fi" << std::endl;
    system( oss.str().c_str() );
    
    //ե뤬¸ߤ뤫¸ߤʤ¸ߤƱΤå
    //exp.conf
    oss.str("");
    oss.clear();
    oss << gconfig.exp().logDir() << "/"
        << gconfig.expConfigFilename();
    std::ifstream ifs( oss.str().c_str() );
    //ե뤬뤫å
    if( ifs ){
        //꤬Ʊå
        if( !gconfig.isSameExpConfig( oss.str() ) ){
            std::cerr << "[ Error] "
                      << gconfig.expConfigFilename()
                      << " is not same! exit." << std::endl;
            safeChangeMode( rcsc::PM_TimeOver);
            return false;
        }
    } else {
        //եʤä饳ԡ
        std::cerr << oss.str() << " not exist. copy file." << std::endl;
        oss.str("");
        oss.clear();
        oss << "cp " << gconfig.expConfigFilename()
            << " " << gconfig.exp().logDir() << std::endl;
        system( oss.str().c_str() );
    }
    ifs.close();
    
    //learn.conf
    oss.str("");
    oss.clear();
    oss << gconfig.exp().logDir() << "/"
        << gconfig.learnConfigFilename();
    ifs.open( oss.str().c_str() );
    //ե뤬뤫å
    if( ifs ){
        //꤬Ʊå
        if( !gconfig.isSameLearnConfig( oss.str() ) ){
            std::cerr << "[ Error] "
                      << gconfig.learnConfigFilename()
                      << " is not same! exit." << std::endl;
            safeChangeMode( rcsc::PM_TimeOver);
            return false;
        }
    } else {
        //եʤä饳ԡ
        std::cerr << oss.str() << " not exist. copy file." << std::endl;
        oss.str("");
        oss.clear();
        oss << "cp " << gconfig.learnConfigFilename()
            << " " << gconfig.exp().logDir() << std::endl;
        system( oss.str().c_str() );
    }
    
    
    //֤ΰ֤
    M_offense_pos.push_back( rcsc::Vector2D(-50,0) );
    M_offense_pos.push_back( rcsc::Vector2D(-25,-5) );
    M_offense_pos.push_back( rcsc::Vector2D(-25,5) );
    M_offense_pos.push_back( rcsc::Vector2D(-25,-10) );
    M_offense_pos.push_back( rcsc::Vector2D(-25,10) );
    M_offense_pos.push_back( rcsc::Vector2D(-25,0) );
    M_offense_pos.push_back( rcsc::Vector2D(-15,-5) );
    M_offense_pos.push_back( rcsc::Vector2D(-15,5) );
    M_offense_pos.push_back( rcsc::Vector2D(-15,-10) );
    M_offense_pos.push_back( rcsc::Vector2D(-15,10) );
    M_offense_pos.push_back( rcsc::Vector2D(-15,0) );

    M_defense_pos.push_back( rcsc::Vector2D(50,0) );
    M_defense_pos.push_back( rcsc::Vector2D(25,5) );
    M_defense_pos.push_back( rcsc::Vector2D(25,-5) );
    M_defense_pos.push_back( rcsc::Vector2D(25,10) );
    M_defense_pos.push_back( rcsc::Vector2D(25,-10) );
    M_defense_pos.push_back( rcsc::Vector2D(25,0) );
    M_defense_pos.push_back( rcsc::Vector2D(15,5) );
    M_defense_pos.push_back( rcsc::Vector2D(15,-5) );
    M_defense_pos.push_back( rcsc::Vector2D(15,10) );
    M_defense_pos.push_back( rcsc::Vector2D(15,-10) );
    M_defense_pos.push_back( rcsc::Vector2D(15,0) );
    
    //֤οå
    if( M_offense_pos.size() != LEFT_PLAYER_NUM ){
        std::cerr << "trainer: Error! rel_offense_pos size is not correct"
                  << std::endl;
        safeChangeMode( rcsc::PM_TimeOver);
        return false;
    }
    if( M_defense_pos.size() != RIGHT_PLAYER_NUM ){
        std::cerr << "trainer: Error! rel_defense_pos size is not correct"
                  << std::endl;
        safeChangeMode( rcsc::PM_TimeOver);
        return false;
    }



    return true;
}

/*-------------------------------------------------------------------*/
/*!

 */
void
SampleTrainer::actionImpl()
{
    //check cycle
    if( world().time() == M_last_decision_time ){
        return;
    } else {
        M_last_decision_time = world().time();
    }
//    if( world().time().cycle() < -10 || world().time().cycle() > 10e5 ){
//        safeChangeMode( rcsc::PM_TimeOver);
//        return;
//    }
    
    //Ϳå
    //ǽ³Ԥ
    static bool S_connect_flg = false;
    static const unsigned int after_connect_wait_time = 10;
    static unsigned int after_connect_time = 0;
    if( !S_connect_flg ){        
        if( world().playersLeft().size() != LEFT_PLAYER_NUM ||
            world().playersRight().size() != RIGHT_PLAYER_NUM ){
            return;  
        } else if( after_connect_time++ < after_connect_wait_time )
            return;
        else{
            //learn_playerɤ߹ߤλ(-20,0)˰ưΤǥå
            bool moved = true;
            rcsc::Vector2D move_pos(-20.0,0);
            GPlayerPtrCont::const_iterator it;
            for(it=world().playersLeft().begin();
                it!=world().playersLeft().end();
                ++it)
            {
                if( (*it)->pos().dist( move_pos ) > 3.0 )
                    moved = false;
                }
            
            if( moved ) S_connect_flg = true;

            return;
        }
    }
    //³Ƥʤå
    if( world().playersLeft().size() != LEFT_PLAYER_NUM ){
        safeChangeMode( rcsc::PM_TimeOver);
        std::cerr << "trainer: **ERROR** playersLeft size is incorrect\n"
                  << "change mode to TimeOver" << std::endl;
        return;
    }
    if( world().playersRight().size() != RIGHT_PLAYER_NUM ){
        std::cerr << "trainer: **ERROR** playersRight size is incorrect\n"
                  << "change mode to TimeOver" << std::endl;
        safeChangeMode( rcsc::PM_TimeOver);
        return;
    }
        
    //ԥɴ֤ϤäԤ
    static const int episode_wait_time = 3;
    if( world().gameMode().type() == rcsc::GameMode::BeforeKickOff
        && world().time().stopped() <= episode_wait_time){
        return;
    }
    
    sampleAction();
}


/*-------------------------------------------------------------------*/
/*!

 */
void
SampleTrainer::sampleAction()
{
    //init
    if( M_init_flg ){
        initPosition();
        doRecover();
        safeChangeMode( rcsc::PM_PlayOn );
        M_init_flg = false;
        return;
    }
    
    if( checkGameMode() ) return;

    if( checkObjectPos() ) return;

}

void
SampleTrainer::initPosition()
{

    const rcsc::Vector2D
        goal_center(rcsc::ServerParam::i().pitchHalfLength(),0);

    //򤳤Ƿ
//    rcsc::UniformReal base_range( -20.0, 20.0 );
//    const rcsc::Vector2D base_pos( 15.0, base_range() );
    
    static const int offense_num = LEFT_PLAYER_NUM;
    static const int defense_num = RIGHT_PLAYER_NUM;
    rcsc::UniformInt baller( 2, offense_num );

    //֤˷
//     RandomGenerator rng;
//     std::vector<int> rand_idx;
//     for(int i=1; i<=offense_num; ++i){
//         rand_idx.push_back(i);
//     }
//     std::random_shuffle(rand_idx.begin(),rand_idx.end(),rng);

    //Offense
    for(int i=1; i<=offense_num; ++i){
//        std::cout << i << " " << M_offense_pos[i-1] << std::endl;
        doMovePlayer(world().teamNameLeft(),i,
                     M_offense_pos[i-1]);
    }

    //֤Τ餻
//    sayLayout(rand_idx);

    //Goalie
//    doMovePlayer(world().teamNameRight(),1,
//                 goal_center + M_rel_defense_pos[0] );

    //Defense
//    doMovePlayer(world().teamNameRight(),2,
//                 goal_center + M_rel_defense_pos[1] );
//    doMovePlayer(world().teamNameRight(),3,
//                 goal_center + M_rel_defense_pos[2] );

    //Defense
    for(int i=1; i<=defense_num; ++i){
//        std::cout << i << " " << M_defense_pos[i-1] << std::endl;
        doMovePlayer(world().teamNameRight(),i,
                     M_defense_pos[i-1]);
    }

    //ball
    int baller_idx = baller()-1;
    doMoveBall(M_offense_pos[baller_idx]
               + (-M_offense_pos[baller_idx]).setLengthVector( 2.0 ),
               rcsc::Vector2D(0,0));
}

bool
SampleTrainer::checkGameMode()
{
    
    //BeforeKickOff
    if( world().gameMode().type() == rcsc::GameMode::BeforeKickOff )
    {
        M_init_flg = true;
        return true;
    }

    //not play on
    if( world().gameMode().type() != rcsc::GameMode::BeforeKickOff
        && world().gameMode().type() != rcsc::GameMode::PlayOn
        && world().gameMode().type() != rcsc::GameMode::TimeOver ){
        std::cerr << world().time()
                  << " **Warning** game mode is neither BeforeKickoff nor "
                  << "PlayOn nor Timeover." << std::endl;
        safeChangeMode( rcsc::PM_BeforeKickOff );
//        M_init_flg = true;
        return true;
    }
    

//     //finish game
//     if( world().time().cycle() >= gconfig.exp().finishTime() ){
//         safeChangeMode( rcsc::PM_TimeOver);
//         return true;
//     }

    return false;
}

bool
SampleTrainer::checkObjectPos()
{
    rcsc::Vector2D ball_pos = world().ball().pos();
    bool episode_terminate = false;

    //check goal
    bool goaled = false;
    if( ball_pos.absY() < rcsc::ServerParam::i().goalHalfWidth()
        && ball_pos.x > rcsc::ServerParam::i().pitchHalfLength() )
    {
        ++M_goal_count;
//        cout << "["<< world().time().cycle() << "] goal!" << endl;

        //say
        doSay( "rg" );
        
        episode_terminate = true;
        goaled = true;
    }

    //check ball pos
    bool inside_field =
        (ball_pos.absY() < rcsc::ServerParam::i().pitchHalfWidth()
         && ball_pos.absX() < rcsc::ServerParam::i().pitchHalfLength()
//         && ball_pos.x > -30
//         && ball_pos.x < rcsc::ServerParam::i().pitchHalfLength()
         );

    if( !inside_field && !goaled ){
//        cout << "["<< world().time().cycle() << "] ball is out of field"
//             << endl;

        //say
        doSay( "ro" );
        
        episode_terminate = true;
    }
    
    //check kickable defense
    bool exist_kickable_right = false;
    bool exist_kickable_right_goalie = false;
    const GPlayerPtrCont & rights = world().playersRight();
    const GPlayerPtrCont::const_iterator rights_end = rights.end();
    for ( GPlayerPtrCont::const_iterator it = rights.begin();
          it != rights_end;
          ++it )
    {
        if( (*it)->goalie() &&
            (*it)->pos().dist( world().ball().pos() )
            < rcsc::ServerParam::i().catchableArea() ){
            exist_kickable_right_goalie = true;
            break;
        }
        
        if( (*it)->pos().dist( world().ball().pos() )
            < rcsc::ServerParam::i().defaultKickableArea() ){
            exist_kickable_right = true;
            break;
        }
    }

    //check kickable offense
    bool exist_kickable_left = false;
    bool exist_kickable_left_goalie = false;
    if( exist_kickable_right ){
        const GPlayerPtrCont & lefts = world().playersLeft();
        const GPlayerPtrCont::const_iterator lefts_end = lefts.end();
        for ( GPlayerPtrCont::const_iterator it = lefts.begin();
              it != lefts_end;
              ++it )
        {
            if( (*it)->goalie() &&
                (*it)->pos().dist( world().ball().pos() )
                < rcsc::ServerParam::i().catchableArea() ){
                exist_kickable_left_goalie = true;
                break;
            }
        
            if( (*it)->pos().dist( world().ball().pos() )
                < rcsc::ServerParam::i().defaultKickableArea() ){
                
                exist_kickable_left = true;
                break;
            }
        }
    }

    //exist catchable goalie
    if( !exist_kickable_left && exist_kickable_right_goalie ){
//        cout << "["<< world().time().cycle()
//             << "] exist catchable goalie" << endl;

        //say
        doSay( "rc" );
        
        episode_terminate = true;
    }

    //exist catchable teammate goalie
    if( !exist_kickable_right && exist_kickable_left_goalie ){

        //say
        doSay( "ro" );
        
        episode_terminate = true;
    }

    //exist kickable defense
    if( !exist_kickable_left && exist_kickable_right ){
//        cout << "["<< world().time().cycle()
//             << "] exist kickable defense" << endl;

        //say
        doSay( "rd" );

        static int hold_time = 0;
        ++hold_time;

        //3Ũݻ齪λ
        if( hold_time >= 3 ){
            episode_terminate = true;
            hold_time = 0;
        }
    }

    if( episode_terminate ){
        ++M_episode_count;
        ++M_episode_sum;

        if( M_episode_count >= log_freq ){
            outputGoalLog();
            M_goal_count = 0;
            M_episode_count = 0;
        }
    
        if( gconfig.exp().finishEpisode() > 0 &&
            M_episode_sum >= gconfig.exp().finishEpisode() ){
            std::cerr << "trainer: its time to finish server"
                      << "change mode to TimeOver" << std::endl;
            safeChangeMode( rcsc::PM_TimeOver);
        } else {
            safeChangeMode( rcsc::PM_BeforeKickOff );
        }
        return true;
    }
    
    return false;
}

bool
SampleTrainer::safeChangeMode( rcsc::PlayMode mode )
{
    bool change_flg = false;
    
    if( M_last_change_mode != mode ){
        doChangeMode( mode );
        change_flg = true;
        M_last_change_mode = mode;
    }

    return change_flg;
}

